<?xml version="1.0" encoding="UTF-8"?>
<xsl:transform version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:um="urn:www.oasis-open.org/ubl/uml-model">
	<xsl:key name="associationEndsByAssociation" match="um:associationEnd" use="um:associationName"/>
	<xsl:template match="@* | node()">
		<xsl:copy>
			<xsl:apply-templates select="@* | node()"/>
		</xsl:copy>
	</xsl:template>
	<xsl:template match="um:UMLModel">
		<xsl:copy>
			<!-- generate associations -->
			<xsl:apply-templates select="//um:associationEnd[generate-id() = generate-id( key('associationEndsByAssociation', um:associationName)[1])]" mode="ASSOC"/>
			<!-- echo the rest of the document -->
			<xsl:apply-templates select="node()"/>
		</xsl:copy>
	</xsl:template>
	<!-- this template will be called once for one (or the other, but not both) end of each association -->
	<xsl:template match="um:associationEnd" mode="ASSOC">
		<xsl:choose>
			<xsl:when test="count( key('associationEndsByAssociation', current()/um:associationName)) = 1">
				<ERROR>Association <xsl:value-of select="current()/um:associationName"/> has only one end: <xsl:call-template name="association-end-identifiers">
						<xsl:with-param name="associationEnds" select="key('associationEndsByAssociation', current()/um:associationName)"/>
					</xsl:call-template>
				</ERROR>
			</xsl:when>
			<xsl:when test="count( key('associationEndsByAssociation', current()/um:associationName)) > 2">
				<ERROR>Association <xsl:value-of select="current()/um:associationName"/> has more than two ends: <xsl:call-template name="association-end-identifiers">
						<xsl:with-param name="associationEnds" select="key('associationEndsByAssociation', current()/um:associationName)"/>
					</xsl:call-template>
				</ERROR>
			</xsl:when>
		</xsl:choose>
		<!-- go ahead and generate the association - even if there were errors -->
		<um:association>
			<um:name>
				<xsl:value-of select="current()/um:associationName"/>
			</um:name>
			<xsl:copy-of select="key('associationEndsByAssociation', current()/um:associationName)"/>
		</um:association>
	</xsl:template>
	<xsl:template name="association-end-identifiers">
		<xsl:param name="associationEnds" select="/.."/>
		<xsl:if test="count($associationEnds)">
			<xsl:text>ABIE=</xsl:text>
			<xsl:value-of select="$associationEnds[1]/../um:name"/>
			<xsl:text> oppositeRoleName=</xsl:text>
			<xsl:value-of select="$associationEnds[1]/um:oppositeRoleName"/>
			<xsl:call-template name="association-end-identifiers">
				<xsl:with-param name="associationEnds" select="associationEnds[position() > 1]"/>
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
</xsl:transform>
