<?xml version="1.0" encoding="UTF-8"?>
<!-- edited with XML Spy v4.3 U (http://www.xmlspy.com) by Bill Burcham (Sterling Commerce) -->
<xsl:transform version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:um="urn:www.oasis-open.org/ubl/uml-model" xmlns:exsl="http://exslt.org/common" xmlns:saxon="http://icl.com/saxon" xmlns:func="http://exslt.org/functions" extension-element-prefixes="exsl saxon func">
	<!--

T h e   L o o p   P r o b l e m
	
The present policy is to elide associationEnds that are caught in loops, i.e. when I see an associationEnd whose oppositeABIEName matches an ABIE in my predecessors I elide that
AssociationEnd.

This predecessors algorithm is a bit strict.  It precludes _recursive_ type structures.  NB: 
recursive structure is not the same as nondeterministic structure - the former is very common
and useful, whereas the latter is less common and arguably less useful (to me anyway ;-)

Now just because we see a predecessor ABIE whose um:name matches the oppositeABIEName
of some AssociationEnd, that doesn't mean that that particular predecessor ABIE variant is 
actually the one we "want".  There are three cases:

A. the predecessor ABIE variant (once it is fully defined) is the one we will want
B. we want to elide the AssociationEnd entirely
C. we want to link to some other variant of the ABIE (other than the predecessor)

The currently policy is to do (B) always.  Supporting (A) would require changes to the
algorithm since with the current algorightm, the predecessor ABIE hasn't really been 
"contextualized" yet so it doesn't havea "variant" name yet so a template couldn't "link" to 
it (even if it wanted to) - recall that the way an AssociationEnd links to an ABIE is via
the pair (um:name um:variantName).  

TODO: In general we should allow loops!  It should be possible for a user to specify that a
a certain association in a certain context should refer "back up" the context.  The way to do this
might be:
1. Add an "internal identifier" (GUID) field to the ABIE.  When processing an ABIE we first
generate and assign that identifier.
2. Use that internal identifier for linking associationEnds to ABIE's (instead of the current scheme
of (um:name, um:variantName).
[3.] any user defined template would be able to "patch" to an association end in the predecessors
by referencing its identifier in an AssociationEnd.
3. (as is currently the case) recursive processing would halt when a loop is detected - only now
the "back reference" would be patched.
4. A separate processing pass should be added after all the ABIE variants have been generated.
This pass replaces the GUID references with more human-readable references (perhaps back
to (um:name, um:variantName).

Once we do this, it should be possible for a user defined template to create candidate ABIEs,
as-yet not fully realized, and refer to them via their internal identifier.  With this, we should be
able to do (C) in a straightforward manner too.

T h e   D o c u m e n t   C o n t e x t   P r o b l e m

We were storing the predecessors structure as a node set.  Each member (peer) was an 
element of the path.  Elements were added at the end.  That wasn't the best structure 
when it came to applying XPath queries for document context.

It'd be ideal if we had a tree-structured rep that we could match on, but building that thing from
the top-down means (costly) copying.  So instead, we build it bottom-up (upside-down), so 
that my nearest predecessors are _below_ me in the infoset (descendant::).  It's a bit counter-intuitive, but let's see how people like it...


TODO: add association end multiplicity constraint checking to the processing.  If the generated
schema allows instances to violate constraints of the base model we need to flag that!
-->
	<xsl:include href="lookup-abie.xslt"/>
	<!--
	Module parameters.  Can be provided on the command line.
	-->
	<xsl:param name="max-depth" select="number(0)"/>
	<xsl:param name="target-namespace" select="'urn:www.oasis-open.org/ubl/2002/11/19/document-model'"/>

	<!--
	Need to capture the root of the input document for later.  Some of the functions and
	templates will be called in a context outside that document and will need to get
	back to it.
	-->
	<xsl:variable name="input-document-root" select="/"/>
	<!--
	This section contains utility templates that provide the processing framework.  These
	utilities are the same for each assembler.
	-->
	<xsl:template match="um:UMLModel">
		<xsl:copy>
			<xsl:copy-of select="@*[name() != 'um:targetNamespace']"/>
			<!-- target namespace id is derived from source namespce and new ns id -->
			<xsl:attribute name="um:targetNamespace"><xsl:value-of select="@um:targetNamespace"/>/<xsl:value-of select="$target-namespace"/></xsl:attribute>
			<xsl:apply-templates select="um:bbie"/>
			<xsl:variable name="hide-output">
				<!-- Delegate to user-defined templates identifying top-level ABIE's-->
				<xsl:apply-templates select="um:abie"/>
			</xsl:variable>
			<!-- generate ABIE variant definitions -->
			<xsl:call-template name="abie-variants"/>
		</xsl:copy>
	</xsl:template>
	<xsl:template match="um:abie">
		<!-- do nothing! -->
	</xsl:template>
	<xsl:template match="um:bbie">
		<xsl:copy-of select="."/>
	</xsl:template>
	<xsl:template match="um:property">
		<xsl:message>PROPERTY RULE</xsl:message>
		<xsl:copy-of select="."/>
	</xsl:template>
	<xsl:template match="um:abieAssemblyContext">
		<xsl:copy-of select="um:abie"/>
	</xsl:template>
	<xsl:template match="um:associationEndAssemblyContext">
		<xsl:copy-of select="um:associationEnd"/>
	</xsl:template>
	<!--
	Returns the generated um:ABIE and the set of um:ABIE variants generated
	thus-far: 

	Returns:	
		<um:abie/>?
	
	-->
	<xsl:template name="process-ABIE">
		<xsl:param name="abie" select="/.."/>
		<xsl:param name="depth" select="number(0)"/>
		<xsl:param name="predecessors" select="/.."/>
		<!--
		<xsl:message>abie(<xsl:value-of select="$abie/um:name"/>,d=<xsl:value-of select="$depth"/>,max=<xsl:value-of select="$max-depth"/>)</xsl:message>
		-->
		<!--
		TODO: construct a naming scheme for output ABIE's. It would be nice if the scheme
		could be computed with local information (e.g. hash) so that we wouldn't have to look
		at all the other ABIE's in order to come up with the name.  Also, we'd like the generated
		name to be meaningful since programmers are going to have to use it.  Also we'd like it 
		to be the case that two destination ABIE's generated from the same source ABIE and 
		having the same structure (and associations) result in the same name.
		Solution: the name of the destination ABIE is the same as the name of the source ABIE
		unless a template acts on the source. In that case, the template must alter the source
		ABIE name in a way that produces a new, meaningful, unique ABIE name.
		With this scheme in place, checking for loops is fairly easy.  We just keep track of the
		path down to this ABIE use, and check that path for this ABIE and if we find it we flag
		it and stop.
		TODO: the algorithm just mentioned isn't perfect. Fix it. Two problems:
		* it's a bit conservative: without the loop checking just described, a user could in general
		write template rules sufficient to break loops.  In that case, the user could allow two 
		destination ABIE's to occur along a path, but could break the circularity with a template.
		It might also be considered a bit conservative since the destination ABIE's generated from 
		different templates but having the same structure will not have the same destination ABIE 
		name.  This could be viewed as a missed opportunity for reuse.  On the other hand, it 
		could be argued that it would be a false reuse since changes to either template could break 
		the structural similarity.
		-->
		<xsl:variable name="assemby-context-rtf">
			<um:abieAssemblyContext>
				<xsl:copy-of select="$abie"/>
				<um:predecessors>
					<xsl:copy-of select="$predecessors"/>
				</um:predecessors>
			</um:abieAssemblyContext>
		</xsl:variable>
		<xsl:variable name="assembly-context" select="exsl:node-set($assemby-context-rtf)"/>
		<xsl:variable name="candidate-ABIE-rtf0">
			<xsl:apply-templates select="$assembly-context"/>
			<!-- <xsl:copy-of select="$abie"/> -->
		</xsl:variable>
		<xsl:variable name="candidate-ABIE-0" select="exsl:node-set($candidate-ABIE-rtf0)"/>
		<xsl:variable name="new-predecessors" select="um:new-predecessors($abie, $predecessors)"/>
		<xsl:variable name="reduced-association-end-set-rtf">
			<xsl:call-template name="remove-cyclical-association-ends">
				<xsl:with-param name="association-end-set" select="$candidate-ABIE-0/um:abie/um:associationEnd"/>
				<xsl:with-param name="predecessors" select="$new-predecessors"/>
			</xsl:call-template>
		</xsl:variable>
		<xsl:variable name="reduced-association-end-set" select="exsl:node-set($reduced-association-end-set-rtf)"/>
		<!-- 
			Construct a new candidate ABIE by removing associationEnds involved in loops.
			-->
		<xsl:variable name="candidate-ABIE-1" select="um:replace-association-ends( $candidate-ABIE-0/um:abie, $reduced-association-end-set/um:associationEnd)"/>
		<xsl:variable name="found" select="um:find-abie($candidate-ABIE-1/um:abie)"/>
		<!--
				Timings.
				Depth Time(user sec)  Notes
				4        50                     Old depth-first algorithm, no pre-selection before search
				3         7                      Pre-select ABIE based on structure (avoid search)
				4         39                     ""
				5         354                   ""
				-->
		<xsl:choose>
			<xsl:when test="count($found) > 0">
				<xsl:message>found BEFORE recursion <xsl:value-of select="um:abie-variant-identifier($found/um:abie)"/>
				</xsl:message>
				<xsl:copy-of select="$found/um:abie"/>
			</xsl:when>
			<xsl:otherwise>
				<xsl:variable name="process-association-end-set-rtf">
					<xsl:apply-templates select="$reduced-association-end-set/um:associationEnd">
						<xsl:with-param name="depth" select="$depth + 1"/>
						<xsl:with-param name="predecessors" select="$new-predecessors"/>
					</xsl:apply-templates>
				</xsl:variable>
				<xsl:variable name="process-association-end-set-result" select="exsl:node-set($process-association-end-set-rtf)"/>
				<!--
				<xsl:message><BOUND><xsl:copy-of select="$process-association-end-set-result"/></BOUND></xsl:message>
				-->
				<xsl:variable name="candidate-ABIE-2" select="um:replace-association-ends( $candidate-ABIE-1/um:abie, $process-association-end-set-result/um:associationEnd)"/>
				<xsl:call-template name="new-abie-variant">
					<xsl:with-param name="candidate-abie-variant" select="$candidate-ABIE-2/um:abie"/>
				</xsl:call-template>
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>
	<!-- 
	Returns the generated um:associationEnd.
	
		<um:associationEnd>...</um:associationEnd>?
	-->
	<xsl:template match="um:associationEnd">
		<xsl:param name="depth" select="number(0)"/>
		<xsl:param name="predecessors" select="/.."/>
		<!--
		<xsl:message>     assoc(<xsl:value-of select="concat(concat(current()/um:oppositeRoleName,':'),current()/um:oppositeABIEName)"/>, d=<xsl:value-of select="$depth"/>)</xsl:message>
		-->
		<xsl:choose>
			<xsl:when test="$depth > $max-depth">
				<!--<xsl:message>depth cut</xsl:message>-->
			</xsl:when>
			<!--
				No need to check for loops here - the caller has already done that.
				-->
			<xsl:otherwise>
				<xsl:variable name="assemby-context-rtf">
					<um:associationEndAssemblyContext>
						<xsl:copy-of select="."/>
						<um:predecessors>
							<xsl:copy-of select="$predecessors"/>
						</um:predecessors>
					</um:associationEndAssemblyContext>
				</xsl:variable>
				<xsl:variable name="assembly-context" select="exsl:node-set($assemby-context-rtf)"/>
				<xsl:variable name="candidate-association-end-rtf">
					<xsl:apply-templates select="$assembly-context"/>
				</xsl:variable>
				<xsl:variable name="candidate-association-end" select="exsl:node-set($candidate-association-end-rtf)"/>
				<!--
				user-defined rule may elide the association end
				-->
				<xsl:if test="count( $candidate-association-end/um:associationEnd) > 0">
					<xsl:variable name="new-predecessors" select="um:new-predecessors( $candidate-association-end/um:associationEnd, $predecessors)"/>
					<!-- TODO: once we allow back-patching, only traverse associations
				that are not already resolved (patched) -->
					<xsl:variable name="dummy" select="um:assert(count($input-document-root//um:abie[um:name = $candidate-association-end/um:associationEnd/um:oppositeABIEName]), 'Cannot find associated ABIE')"/>
					<xsl:variable name="process-ABIE-rtf">
						<xsl:call-template name="process-ABIE">
							<xsl:with-param name="abie" select="$input-document-root//um:abie[um:name = $candidate-association-end/um:associationEnd/um:oppositeABIEName]"/>
							<xsl:with-param name="depth" select="$depth"/>
							<xsl:with-param name="predecessors" select="$new-predecessors"/>
						</xsl:call-template>
					</xsl:variable>
					<xsl:variable name="process-ABIE-result" select="exsl:node-set($process-ABIE-rtf)"/>
					<xsl:copy>
						<xsl:copy-of select="*[name() != 'um:oppositeABIEName' and name() != 'um:oppositeABIEVariantName']"/>
						<um:oppositeABIEName>
							<xsl:value-of select="$process-ABIE-result/um:abie/um:name"/>
						</um:oppositeABIEName>
						<um:oppositeABIEVariantName>
							<xsl:value-of select="$process-ABIE-result/um:abie/um:variantName"/>
						</um:oppositeABIEVariantName>
					</xsl:copy>
					<!--
					<xsl:message>     assoc-end(<xsl:value-of select="$association-end/um:oppositeRoleName"/>.<xsl:value-of select="$process-ABIE-result/process-abie-result/abie/um:abie/um:variantName"/>, d=<xsl:value-of select="$depth"/>)</xsl:message>
					-->
				</xsl:if>
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>
	<!-- 
	Given a set of um:associationEnd and a predecessors tree, eliminate those 
	um:associationEnds that are represented in the predecessors tree.
	
	Returns:
		<um:associationEnd/>*
		
	TODO: allow cyclical ends that have been "patched".  Do this once we have the two-pass
	algorithm in place.
	-->
	<xsl:template name="remove-cyclical-association-ends">
		<xsl:param name="association-end-set" select="/.."/>
		<xsl:param name="predecessors" select="/.."/>
		<xsl:if test="count($association-end-set) > 0">
			<xsl:if test="count( $predecessors/descendant-or-self::node()[ name(.) = 'um:abie' and um:name = $association-end-set[1]/um:oppositeABIEName] ) &lt; 1">
				<!--
				<xsl:message><COPYING></COPYING></xsl:message>
				-->
				<xsl:copy-of select="$association-end-set[1]"/>
			</xsl:if>
			<xsl:call-template name="remove-cyclical-association-ends">
				<xsl:with-param name="association-end-set" select="$association-end-set[ position() > 1]"/>
				<xsl:with-param name="predecessors" select="$predecessors"/>
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	<!--
	Build a new predecessors tree.  Returns a node set - not a result tree fragment.
	The predecessors tree has a branching factor of 1 and is upside-down.  It's the best we
	can do.  Building it right-side-up would entail extending XSLT to allow us to have
	two _input_ documents at once.   I don't know how to do that.  The outcome is that
	user templates will look a little funny.  They'll call out descendant:: in the predecessors
	to mean ancestor:: in the "document context".  Ugh.  Sorry.
	
	The actual predecessors tree is made up of nested um:abie and um:associationEnd nodes,
	alternating.  This means that the nodes in this tree aren't _really_ um:abie or um:associationEnd
	nodes per the um schema... they're tweaked a little so that they nest.  I could have defined
	a whole new type of node for this, but I think doing it this way makes the meaning more 
	apparent to user templates.
	-->
	<func:function name="um:new-predecessors">
		<xsl:param name="new-child" select="/.."/>
		<xsl:param name="old-predecessors" select="/.."/>
		<xsl:variable name="result-rtf">
			<xsl:for-each select="$new-child">
				<!-- build the (inverted) tree of predecessors for document context -->
				<xsl:copy>
					<xsl:copy-of select="@* | node()"/>
					<xsl:copy-of select="$old-predecessors"/>
				</xsl:copy>
			</xsl:for-each>
		</xsl:variable>
		<func:result select="exsl:node-set($result-rtf)"/>
	</func:function>
	<func:function name="um:assert">
		<xsl:param name="tst" select="0"/>
		<xsl:param name="msg" select="''"/>
		<xsl:if test="not( $tst)">
			<xsl:message>
				<xsl:value-of select="$msg"/>
			</xsl:message>
		</xsl:if>
		<func:result select="''"/>
	</func:function>
	<func:function name="um:replace-association-ends">
		<xsl:param name="abie" select="/.."/>
		<xsl:param name="new-association-ends" select="/.."/>
		<xsl:for-each select="$abie">
			<xsl:variable name="new-abie-rtf">
				<xsl:copy>
					<xsl:copy-of select="@* | node()[name() != 'um:associationEnd']"/>
					<!--					<xsl:copy-of select="$new-association-ends/node()"/>-->
					<xsl:copy-of select="$new-association-ends"/>
				</xsl:copy>
			</xsl:variable>
			<func:result select="exsl:node-set( $new-abie-rtf)"/>
		</xsl:for-each>
	</func:function>
</xsl:transform>
