<?xml version="1.0" encoding="UTF-8"?>
<xsl:transform version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:um="urn:www.oasis-open.org/ubl/uml-model" xmlns:exsl="http://exslt.org/common" xmlns:saxon="http://icl.com/saxon" xmlns:func="http://exslt.org/functions" extension-element-prefixes="exsl saxon func">
	<!--
	T h e   P e r f o r m a n c e   P r o b l e m   W i t h   T h e   V a r i a n t s   S e t

The set of ABIE variants is essentially a global stucture, but to avoid having a global variable
to hold it, I initially passed it around _everywhere_.  This caused horrendous performance
problems in Saxon.   Now we use saxon:assign and a global variable to hold that thing so that
we only (possibly) copy it when we define a new variant - not every time we make a call!
	-->
	<!--
	the global set of ABIE variants
		
	<abie-variants>
		<um:abie>...</um:abie>+
	</abie-variants>
	-->
	<xsl:variable name="abie-variants" select="/.." saxon:assignable="yes"/>
	<xsl:template name="abie-variants">
		<xsl:copy-of select="$abie-variants"/>
	</xsl:template>
	<!--
	Return the human-readable variant identifier.
	-->
	<func:function name="um:abie-variant-identifier">
		<xsl:param name="abie-variant" select="/.."/>
		<xsl:choose>
			<xsl:when test="string-length( $abie-variant/um:variantName) = 0">
				<func:result select="$abie-variant/um:name"/>
			</xsl:when>
			<xsl:otherwise>
				<func:result select="concat(concat($abie-variant/um:name,'.'),$abie-variant/um:variantName)"/>
			</xsl:otherwise>
		</xsl:choose>
	</func:function>
	<!--
	Finds a matching ABIE variant.  The matching criteria ignores the um:variantName element of 
	um:abie, and the um:oppositeABIEVariantName element of um:associationEnd.
	
	This routing is used to find an ABIE variant suitable for replacing a (plain) ABIE.
	
	Returns an empty result tree fragment if no suitable ABIE variant is found.
	-->
	<func:function name="um:find-abie">
		<xsl:param name="abie" select="/.."/>
		<func:result select="$abie-variants[um:abie-content-structure(um:abie) = um:abie-content-structure($abie)]"/>
	</func:function>
	<!--
	Finds another variant that differs at most in the um:variantName.
	-->
	<func:function name="um:find-abie-variant">
		<xsl:param name="abie-variant" select="/.."/>
		<func:result select="$abie-variants[um:abie-variant-content-structure(um:abie) = um:abie-variant-content-structure($abie-variant)]"/>
	</func:function>
	<!--
Finds a matching ABIE variant and returns it if one exists.
Otherwise, adds the new ABIE to the set and returns the new variant.
Returns:

	<um:abie>...</um:abie>

Alters global abie-variants.
	-->
	<xsl:template name="new-abie-variant">
		<xsl:param name="candidate-abie-variant" select="/.."/>
		<xsl:variable name="found-variant" select="$abie-variants[um:abie-variant-content-structure(um:abie) = um:abie-variant-content-structure($candidate-abie-variant)]"/>
		<xsl:choose>
			<xsl:when test="count($found-variant)">
				<xsl:message>REUSED variant <xsl:value-of select="um:abie-variant-identifier($found-variant/um:abie)"/>
				</xsl:message>
				<xsl:copy-of select="$found-variant"/>
			</xsl:when>
			<xsl:otherwise>
				<xsl:variable name="new-variant-rtf">
					<um:abie>
						<!-- per the schema, name, variantName come before other elements -->
						<xsl:copy-of select="$candidate-abie-variant/um:name"/>
						<um:variantName>
							<xsl:value-of select="count($abie-variants[um:abie/um:name = $candidate-abie-variant/um:name])"/>
						</um:variantName>
						<xsl:copy-of select="$candidate-abie-variant/node()[name() != 'um:variantName' and name() != 'um:name']"/>
					</um:abie>
				</xsl:variable>
				<xsl:variable name="new-variant" select="exsl:node-set($new-variant-rtf)"/>
				<xsl:message>CREATED variant <xsl:value-of select="um:abie-variant-identifier($new-variant/um:abie)"/>
				</xsl:message>
				<xsl:copy-of select="$new-variant"/>
				<saxon:assign name="abie-variants" select="$abie-variants | $new-variant"/>
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>
	<func:function name="um:abie-variant-content-structure">
		<xsl:param name="arg" select="."/>
		<!--
		CAUTION: at least in SAXON, turning on these messages causes their output
		to pollute the calling XPath!!!!!
		-->
		<!--
		<xsl:message> EVALUATING <xsl:value-of select="$arg/um:name"/>.<xsl:value-of select="$arg/um:variantName"/> (<xsl:copy-of select="$arg"/>)</xsl:message>
		-->
		<xsl:variable name="canonical-rtf">
			<node>
				<!-- The canonical form excludes the um:variantName element -->
				<xsl:copy-of select="$arg/@* | $arg/node()[name() != 'um:variantName']"/>
			</node>
		</xsl:variable>
		<xsl:variable name="canonical" select="exsl:node-set($canonical-rtf)"/>
		<func:result>
			<!--
			TODO: this "fingerprint" might not be sufficient.  Since it simply concatenates the values
			of all the nodes there is the possibility that to different structures will have the same
			"fingerprint".  Think about that some more.
			-->
			<xsl:value-of select="$canonical"/>
			<!--
			<xsl:message>  CRITERIA <xsl:value-of select="$canonical"/>(<xsl:copy-of select="$canonical"/>)
			</xsl:message>
			-->
		</func:result>
	</func:function>
	<func:function name="um:abie-content-structure">
		<xsl:param name="arg" select="."/>
		<xsl:variable name="canonical-rtf">
			<xsl:for-each select="$arg">
				<xsl:copy>
					<!-- The canonical form excludes the um:variantName element -->
					<xsl:copy-of select="@* | node()[name() != 'um:variantName' and name() != 'um:associationEnd']"/>
					<xsl:for-each select="um:associationEnd">
						<xsl:copy>
							<!-- The canonical form excludes the um:oppositeABIEVariantName -->
							<xsl:copy-of select="@* | node()[name() != 'um:oppositeABIEVariantName']"/>
						</xsl:copy>
					</xsl:for-each>
				</xsl:copy>
			</xsl:for-each>
		</xsl:variable>
		<xsl:variable name="canonical" select="exsl:node-set($canonical-rtf)"/>
		<func:result>
			<!--
			TODO: this "fingerprint" might not be sufficient.  Since it simply concatenates the values
			of all the nodes there is the possibility that to different structures will have the same
			"fingerprint".  Think about that some more.
			-->
			<xsl:value-of select="$canonical"/>
		</func:result>
	</func:function>
</xsl:transform>
