<?xml version="1.0" encoding="UTF-8"?>
<xsl:transform version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:ss="urn:schemas-microsoft-com:office:spreadsheet" xmlns:um="urn:www.oasis-open.org/ubl/uml-model" xmlns:func="http://exslt.org/functions" extension-element-prefixes="func" exclude-result-prefixes="um ss">
	<xsl:include href="normalize-identifier.xslt"/>
	<xsl:param name="target-namespace" select="'urn:www.oasis-open.org/ubl/2002/11/19/uml-model'"/>
	<xsl:template match="/">
		<!--
	I'd like to use a literal result element here, but unfortunately doing so causes a validation
	error in this transform.  Using CDATA causes the result document to be not well formed.
	So I'm building this stinking element the hard way
	-->
		<xsl:element name="um:UMLModel" namespace="urn:www.oasis-open.org/ubl/uml-model" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
			<xsl:attribute name="xsi:schemaLocation">urn:www.oasis-open.org/ubl/uml-model ../UMLModel.xsd</xsl:attribute>
			<xsl:attribute name="um:targetNamespace"><xsl:value-of select="$target-namespace"/></xsl:attribute>
			<!-- Generate (unique!!) BBIE definitions -->
			<xsl:for-each select="//ss:Row[ss:Cell[11]/ss:Data = 'BBIE']/ss:Cell[5]/ss:Data[not(.=../../preceding-sibling::ss:Row/ss:Cell[5]/ss:Data)]">
				<um:bbie>
					<um:name>
						<xsl:value-of select="um:normalize-identifier(current())"/>
					</um:name>
					<um:cctName>
						<xsl:value-of select="um:normalize-identifier(current())"/>
					</um:cctName>
				</um:bbie>
			</xsl:for-each>
			<!-- Generate ABIE definitions -->
			<xsl:call-template name="process-rows">
				<xsl:with-param name="rows" select="//ss:Row"/>
			</xsl:call-template>
		</xsl:element>
	</xsl:template>
	<xsl:template name="process-rows">
		<xsl:param name="rows" select="/.."/>
		<xsl:if test="$rows">
			<xsl:if test="$rows[1]/ss:Cell[11]/ss:Data = 'ABIE'">
				<xsl:call-template name="process-ABIE">
					<xsl:with-param name="rows" select="$rows"/>
				</xsl:call-template>
			</xsl:if>
			<xsl:call-template name="process-rows">
				<xsl:with-param name="rows" select="$rows[position() > 1]"/>
			</xsl:call-template>
		</xsl:if>
	</xsl:template>
	<xsl:template name="process-ABIE">
		<xsl:param name="rows" select="/.."/>
		<um:abie>
			<um:name>
				<xsl:value-of select="um:normalize-identifier($rows[1]/ss:Cell[1]/ss:Data)"/>
			</um:name>
			<um:variantName>
				<!-- empty! -->
			</um:variantName>
			<xsl:call-template name="process-ABIE-members">
				<xsl:with-param name="rows" select="$rows[position() > 1]"/>
			</xsl:call-template>
		</um:abie>
	</xsl:template>
	<xsl:template name="process-ABIE-members">
		<xsl:param name="rows" select="/.."/>
		<xsl:if test="$rows">
			<xsl:choose>
				<xsl:when test="$rows[1]/ss:Cell[11]/ss:Data = 'BBIE'">
					<um:property>
						<um:name>
							<xsl:value-of select="um:normalize-identifier(concat($rows[1]/ss:Cell[2]/ss:Data, $rows[1]/ss:Cell[3]/ss:Data))"/>
						</um:name>
						<um:oppositeBBIEName>
							<xsl:value-of select="um:normalize-identifier($rows[1]/ss:Cell[5]/ss:Data)"/>
						</um:oppositeBBIEName>
						<um:oppositeMultiplicity>
							<um:minimum>
								<xsl:value-of select="substring-before($rows[1]/ss:Cell[10]/ss:Data, '..')"/>
							</um:minimum>
							<um:maximum>
								<xsl:value-of select="substring-after($rows[1]/ss:Cell[10]/ss:Data, '..')"/>
							</um:maximum>
						</um:oppositeMultiplicity>
					</um:property>
					<xsl:call-template name="process-ABIE-members">
						<xsl:with-param name="rows" select="$rows[position() > 1]"/>
					</xsl:call-template>
				</xsl:when>
				<xsl:when test="$rows[1]/ss:Cell[11]/ss:Data = 'ASBIE'">
					<um:associationEnd>
						<um:oppositeRoleName>
							<xsl:value-of select="um:normalize-identifier(concat($rows[1]/ss:Cell[7]/ss:Data, $rows[1]/ss:Cell[8]/ss:Data))"/>
						</um:oppositeRoleName>
						<um:oppositeABIEName>
							<xsl:value-of select="um:normalize-identifier($rows[1]/ss:Cell[8]/ss:Data)"/>
						</um:oppositeABIEName>
						<um:oppositeABIEVariantName>
							<!-- empty -->
						</um:oppositeABIEVariantName>
						<um:oppositeMultiplicity>
							<um:minimum>
								<xsl:value-of select="substring-before($rows[1]/ss:Cell[10]/ss:Data, '..')"/>
							</um:minimum>
							<um:maximum>
								<xsl:value-of select="substring-after($rows[1]/ss:Cell[10]/ss:Data, '..')"/>
							</um:maximum>
						</um:oppositeMultiplicity>
						<um:associationName>
						<!-- next transform in the pipeline will fill this in -->
						</um:associationName>
					</um:associationEnd>
					<xsl:call-template name="process-ABIE-members">
						<xsl:with-param name="rows" select="$rows[position() > 1]"/>
					</xsl:call-template>
				</xsl:when>
			</xsl:choose>
		</xsl:if>
	</xsl:template>
</xsl:transform>
